Attribute VB_Name = "modMatrici"
'========================================================
' Descrizione.....: Routines per calcolo matriciale.
' Nome dei Files..: Matrici.bas
' Data............: 11/2/2000
' Aggiornamento...: 12/1/2001
' Versione........: 1.01 a 32 bits.
' Sistema.........: VB6 sotto Windows NT.
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://members.xoom.it/flanguasco/
'                   http://www.mclink.it/personal/MC7061/
'========================================================
'   Comprende le seguenti funzioni e subroutines:
'
'    c() = MSom(a(), b()):            ritorna la somma di due matrici.
'    c() = MDif(a(), b()):            ritorna la differenza di due matrici.
'    c() = MMol(a(), b()):            ritorna il prodotto di due matrici.
'    c() = MMSc(a(),s):               ritorna il prodotto di una matrice per uno scalare.
'    c() = MIde(N):                   ritorna una matrice identica di ordine N.
'    c() = Trasposta(a()):            ritorna la trasposta di una matrice.
'    B = MEgu(a(), b()):              verifica l' eguaglianza a() = b().
'    B = Inversa(a(), c()):           calcola la matrice inversa (si)(*).
'    B = Determinante(a(), det):      calcola il determinante di una matrice
'                                     quadrata (si)(*).
'
'    B = LinEqu(a(), b()):            calcola le soluzioni di un sistema di
'                                     equazioni lineari (si)(*).
'    GAUS a(), b(), x(), IERROR:      calcola le soluzioni di un sistema di
'                                     equazioni lineari (**).
'    TRISOL a(), b(), c(), d():       risolve un sistema di equazioni tridiagonale
'                                     con il metodo di Thomas.
'    B = Jacobi (a(), d(), v()):      calcola gli autovalori e gli autovettori di una
'                                     matrice quadrata, reale e simmetrica (*).
'    B = AutoValori(a(), wr(), wi()): calcola gli autovalori, reali e complessi, di una
'                                     matrice quadrata e reale (*).
'    POWERM a(), v(), lambda:         calcola l' autovalore dominante e l' autovettore
'                                     corrispondente, reali, di una matrice quadrata e
'                                     reale (**).
'    POWINV a(), v(), alfa, lambda:   calcola l' autovalore piu' vicino ad alfa ed il
'                                     corrispondente autovettore, reali, di una matrice
'                                     quadrata e reale (**).
'
'   (si):   ritornano False se la matrice e' singolare.
'           Vedere l' ultima istruzione in ludcmp.
'
'   Tipi:   a() as Double, b() as Double, c() as Double, d() as Double
'           v(), wr() as Double, wi() as Double
'           d as Double, det as Double, s as Double, alfa as Double, lambda as Double
'           N as Long, IERROR as Long
'           B as Boolean
'
'   Note di traduzione: le matrici ed i vettori marcati ByRef nella dichiarazione
'                       iniziale delle routines vengono modificati dalle stesse.
'                       E' stata mantenuta, per quanto possibile, la struttura dei
'                       programmi originali: ne deriva l' uso intensivo dei GoTo
'                       (cosi' invisi ai puristi...).
'
'   Nota:   tutte le matrici ed i vettori usati da queste  routines
'           iniziano dall' indice 1: e.g. a(1 To N, 1 To N)
'
'   (*):    tratte da:  Numerical Recipes in Fortran 77.
'                       The Art of Scientific Computing.
'                       Second Edition.
'                       Volume 1 of Fortran Numerical Recipes.
'                       William H. Press, Saul A. Teukolsky,
'                       William T. Vetterling, Brian P. Flannery
'
'   (**):   tratte da:  Numerical Analysis.
'                       L. W. Johnson, R. D. Riess
'                       Addison-Wesley Publishing Company
'
Option Explicit
Private Sub balanc(ByRef a#())
'
'   Given an N by N matrix a() stored in an array, this routine replaces
'   it by a balanced matrix with identical eigenvalues.
'   A symmetric matrix is already balanced and is unaffected
'   by this procedure. The parameter RADIX should be the machine's
'   floating-point radix.
'
    Dim I&, J&, N&, Last&
    Dim c#, f#, g#, r#, s#
    Const RADIX# = 2#, SQRDX# = RADIX * RADIX
'
    N = UBound(a, 1)
'
1
    Last = 1
    For I = 1 To N  ' Calculate row and column norms.
        c = 0#
        r = 0#
        For J = 1 To N
            If (J <> I) Then
                c = c + Abs(a(J, I))
                r = r + Abs(a(I, J))
            End If
        Next J
        If (c <> 0# And r <> 0#) Then ' If both are nonzero,
            g = r / RADIX
            f = 1#
            s = c + r
'
2
            If (c < g) Then ' find the integer power of the machine radix that
                            ' comes closest to balancing the matrix.
                f = f * RADIX
                c = c * SQRDX
                GoTo 2
            End If
            g = r * RADIX
'
3
            If (c > g) Then
                f = f / RADIX
                c = c / SQRDX
                GoTo 3
            End If
'
            If ((c + r) / f < 0.95 * s) Then
                Last = 0
                g = 1# / f
                For J = 1 To N ' Apply similarity transformation.
                    a(I, J) = a(I, J) * g
                Next J
                For J = 1 To N
                    a(J, I) = a(J, I) * f
                Next J
            End If
        End If
    Next I
    If (Last = 0) Then GoTo 1
'
'
'
End Sub
Public Sub TRISOL(a#(), ByRef b#(), c#(), ByRef d#())
'
'   Risolve un sistema di equazioni tridiagonale
'   con il metodo di Thomas.
'    a(1 To N): diagonale bassa.
'    b(1 To N): diagonale centrale.
'    c(1 To N): diagonale alta.
'    d(1 To N): entra con il vettore dei termini noti;
'               esce con il vettore delle soluzioni.
'   Deve essere:
'    a(1) = 0.
'    c(N) = 0.
'
'   Da: Finite Difference Methods in Heat Transfer
'       M. Necati Ozisik
'       CRC Press
'
    Dim I&, J&, N&, r#
'
    N = UBound(a)
'
    For I = 2 To N
        r = a(I) / b(I - 1)
        b(I) = b(I) - r * c(I - 1)
        d(I) = d(I) - r * d(I - 1)
    Next I
'
    d(N) = d(N) / b(N)
    For I = 2 To N
        J = N - I + 1
        d(J) = (d(J) - c(J) * d(J + 1)) / b(J)
    Next I
'
'
'
End Sub

Public Function Jacobi(ByRef a#(), ByRef d#(), ByRef v#() _
    , Optional ByRef NRot&) As Boolean
'
'   Computes all eigenvalues and eigenvectors of a real symmetric matrix a(),
'   which is of size N by N. Only the elements of a() above the diagonal are cosidered.
'   On output, elements of a() above the diagonal are destroyed.
'   d() returns the eigenvalues of a() in its first N elements.
'   v() is a matrix with the same logical and physical dimensions as a(),
'   whose columns contain, on output, the normalized eigenvectors of a().
'   NRot returns the number of Jacobi rotations that were required.
'
    Dim I&, IP&, IQ&, J&, N&
    Dim c#, g#, h#, s#, sm#, t#, tau#, theta#, tresh#
'
    N = UBound(a, 1)
    ReDim d(1 To N), v(1 To N, 1 To N)
    ReDim b#(1 To N), z#(1 To N)
'
    For IP = 1 To N         ' Initialize to the identity matrix.
        For IQ = 1 To N
            v(IP, IQ) = 0#
        Next IQ
        v(IP, IP) = 1#
    Next IP
'
    For IP = 1 To N
        b(IP) = a(IP, IP)   ' Initialize b and d to the diagonal of a.
        d(IP) = b(IP)
        z(IP) = 0#          ' This vector will accumulate terms of the form TApq
                            ' as in equation (11.1.14).
    Next IP
'
    NRot = 0
    For I = 1 To 50
        sm = 0#
        For IP = 1 To N - 1 ' Sum off-diagonal elements.
            For IQ = IP + 1 To N
                sm = sm + Abs(a(IP, IQ))
            Next IQ
        Next IP
        'If (sm = 0#) Then
        If (Abs(sm) < 1E-20) Then
            Jacobi = True   ' The normal return, which relies on
            Exit Function   ' quadratic convergence to machine underflow.
        End If
'
        If (I < 4) Then
            tresh = 0.2 * sm / (N ^ 2)  ' ...on the first three sweeps.
        Else
            tresh = 0#                  ' ...thereafter.
        End If
'
        For IP = 1 To N - 1
            For IQ = IP + 1 To N
                g = 100# * Abs(a(IP, IQ))
                ' After four sweeps, skip the rotation if the
                ' off-diagonal element is small:
                If ((I > 4) And ((Abs(d(IP)) + g) = Abs(d(IP))) _
                And ((Abs(d(IQ)) + g) = Abs(d(IQ)))) Then
                    a(IP, IQ) = 0#
                ElseIf (Abs(a(IP, IQ)) > tresh) Then
                    h = d(IQ) - d(IP)
                    If (Abs(h) + g = Abs(h)) Then
                        t = a(IP, IQ) / h
                    Else
                        theta = 0.5 * h / a(IP, IQ) ' Equation (11.1.10).
                        t = 1# / (Abs(theta) + Sqr(1# + theta ^ 2))
                        If (theta < 0#) Then t = -t
                    End If
                    c = 1# / Sqr(1 + t ^ 2)
                    s = t * c
                    tau = s / (1# + c)
                    h = t * a(IP, IQ)
                    z(IP) = z(IP) - h
                    z(IQ) = z(IQ) + h
                    d(IP) = d(IP) - h
                    d(IQ) = d(IQ) + h
                    a(IP, IQ) = 0#
                    For J = 1 To IP - 1         ' Case of rotations 1 <= j < p.
                        g = a(J, IP)
                        h = a(J, IQ)
                        a(J, IP) = g - s * (h + g * tau)
                        a(J, IQ) = h + s * (g - h * tau)
                    Next J
                    For J = IP + 1 To IQ - 1    ' Case of rotations p < j < q.
                        g = a(IP, J)
                        h = a(J, IQ)
                        a(IP, J) = g - s * (h + g * tau)
                        a(J, IQ) = h + s * (g - h * tau)
                    Next J
                    For J = IQ + 1 To N         ' Case of rotations q < j <= n.
                        g = a(IP, J)
                        h = a(IQ, J)
                        a(IP, J) = g - s * (h + g * tau)
                        a(IQ, J) = h + s * (g - h * tau)
                    Next J
                    For J = 1 To N
                        g = v(J, IP)
                        h = v(J, IQ)
                        v(J, IP) = g - s * (h + g * tau)
                        v(J, IQ) = h + s * (g - h * tau)
                    Next J
                    NRot = NRot + 1
                End If
            Next IQ
        Next IP
'
        For IP = 1 To N
            b(IP) = b(IP) + z(IP)
            d(IP) = b(IP)   ' Update d with the sum of tapq,
            z(IP) = 0#      ' and reinitialize z.
        Next IP
    Next I
'
'
Jacobi_ERR:
    Jacobi = False
    MsgBox "Troppe iterazioni.", vbCritical, " Jacobi"
'
'
'
End Function


Private Function ludcmp(ByRef a#(), ByRef Indx&(), ByRef d#) As Boolean
'
'   Given a matrix a(1 To N, 1 To N), this routine replaces it by
'   the LU decomposition of a rowwise permutation of itself.
'   a() is the input matrix. a() is output, arranged as in equation
'   (2.3.14) above; Indx(1 To N) is an output vector that records the
'   row permutation effected by the partial pivoting;
'   d is output as  1 depending on whether the number of row interchanges
'   was even or odd, respectively. This routine is used in combination
'   with lubksb to solve linear equations or invert a matrix.
'
    Dim I&, IMax&, N&, J&, K&, bSing As Boolean
    Dim aamax#, dum#, sum#, vv#()
    Const tiny# = 1E-20     ' A small number.
'
    N = UBound(a, 1)
    ReDim Indx(1 To N)
    ReDim vv#(1 To N)       ' vv stores the implicit scaling of each row.
    bSing = False
'
    d = 1#                  ' No row interchanges yet.
    For I = 1 To N          ' Loop over rows to get the implicit scaling information.
        aamax = 0#
        For J = 1 To N
            If (Abs(a(I, J)) > aamax) Then aamax = Abs(a(I, J))
        Next J
        If (aamax = 0#) Then
            MsgBox "Singular matrix in ludcmp." & vbNewLine _
                 & "No nonzero largest element.", vbCritical, " ludcmp"
            ludcmp = False
            Exit Function
        End If
        vv(I) = 1# / aamax  ' Save the scaling.
    Next I
'
    For J = 1 To N          ' This is the loop over columns of Crout's method.
        For I = 1 To J - 1  ' This is equation (2.3.12) except for i = j.
            sum = a(I, J)
            For K = 1 To I - 1
                sum = sum - a(I, K) * a(K, J)
            Next K
            a(I, J) = sum
        Next I
        aamax = 0#          ' Initialize for the search for largest pivot element.
        For I = J To N      ' This is i = j of equation (2.3.12) and i = j +1 ... N
                            ' of equation (2.3.13).
            sum = a(I, J)
            For K = 1 To J - 1
                sum = sum - a(I, K) * a(K, J)
            Next K
            a(I, J) = sum
            dum = vv(I) * Abs(sum)  ' Figure of merit for the pivot.
            If (dum >= aamax) Then  ' Is it better than the best so far?
                IMax = I
                aamax = dum
            End If
        Next I
'
        If (J <> IMax) Then         ' Do we need to interchange rows?
            For K = 1 To N          ' Yes, do so...
                dum = a(IMax, K)
                a(IMax, K) = a(J, K)
                a(J, K) = dum
            Next K
            d = -d                  ' ...and change the parity of d.
            vv(IMax) = vv(J)        ' Also interchange the scale factor.
        End If
        Indx(J) = IMax
1       If (a(J, J) = 0#) Then      ' If the pivot element is zero the matrix is singular
            a(J, J) = tiny          ' (at least to the precision of the algorithm).
                                    ' For some applications on singular matrices, it is
                                    ' desirable to substitute TINY for zero.
            bSing = True
        End If
'
        If (J <> N) Then            ' Now, finally, divide by the pivot element.
            dum = 1# / a(J, J)
            For I = J + 1 To N
                a(I, J) = a(I, J) * dum
            Next I
        End If
    Next J                          ' Go back for the next column in the reduction.
'
    ludcmp = True
    ludcmp = ludcmp And (Not bSing) ' Opzionale: per avvertire della singolarita'.
                                    ' Vedi commento originale all' istruzione 1
                                    ' e sg.
'
'
'
End Function

Private Sub lubksb(a#(), Indx&(), ByRef b#())
'
'   Solves the set of N linear equations A * X = B.
'   Here a(1 To N, 1 To N) is input, not as the matrix A but rather
'   as its LU decomposition, determined by the routine ludcmp.
'   Indx(1 To N) is input as the permutation vector returned by ludcmp.
'   b(1 To N) is input as the right-hand side vector B, and returns
'   with the solution vector X.
'   a() and Indx() are not modified by this routine and can be left
'   in place for successive calls with different right-hand sides b().
'   This routine takes into account the possibility that b() will begin
'   with many zero elements, so it is efficient for use in matrix inversion.
'
    Dim I&, N&, II&, J&, LL&
    Dim sum#
'
    N = UBound(a, 1)
'
    II = 0              ' When II is set to a positive value, it will become the in-
                        ' dex of the first nonvanishing element of b(). We now do
                        ' the forward substitution, equation (2.3.6). The only new
                        ' wrinkle is to unscramble the permutation as we go.
    For I = 1 To N
        LL = Indx(I)
        sum = b(LL)
        b(LL) = b(I)
        If (II <> 0) Then
            For J = II To I - 1
                sum = sum - a(I, J) * b(J)
            Next J
        ElseIf (sum <> 0#) Then
            II = I      ' A nonzero element was encountered, so from now on we will
                        ' have to do the sums in the loop above.
        End If
        b(I) = sum
    Next I
'
    For I = N To 1 Step -1      ' Now we do the backsubstitution, equation (2.3.7).
        sum = b(I)
        For J = I + 1 To N
            sum = sum - a(I, J) * b(J)
        Next J
        b(I) = sum / a(I, I)    ' Store a component of the solution vector X.
    Next I
'
'
'
End Sub

Public Function Inversa_2(ByRef a() As Double, ByRef c() As Double _
    , Optional ByRef det As Double) As Boolean
'
'   Ritorna in c() l' inversa della matrice quadrata a()
'   ed in det il determinante di a().
'   La matrice originale a() viene modificata.
'   Se durante il calcolo vengono trovate condizioni
'   anomale la funzione ritorna False.
'   Nota: l' algoritmo usato non e' molto robusto:
'         per matrici molto grandi e' preferibile
'         usare la routine Inversa.
'
'   Da: Basic Computer Programs in Science and Engineering
'       J. H. Gilder
'       Hayden Book Company.
'
    Dim I As Long, J As Long, L As Long, K As Long, N As Long
    Dim t As Double, s As Double, d As Double
'
    On Error GoTo Inversa_2_ERR
'
    N = UBound(a, 1)
    ReDim c(1 To N, 1 To N) As Double
'
    For I = 1 To N
        For J = 1 To N
            c(I, J) = 0#
        Next J
        c(I, I) = 1#
    Next I
'
    det = 1#
    For J = 1 To N
        For I = J To N
            If a(I, J) <> 0 Then Exit For
        Next I
        If I > N Then Err.Raise 5, , "La matrice e' singolare."
'
        For L = 1 To N
            t = a(J, L)
            a(J, L) = a(I, L)
            a(I, L) = t
            s = c(J, L)
            c(J, L) = c(I, L)
            c(I, L) = s
        Next L
'
        d = a(J, J)
        det = det * d
        For L = 1 To N
            a(J, L) = a(J, L) / d
            c(J, L) = c(J, L) / d
        Next L
'
        For L = 1 To N
            If J <> L Then
                d = a(L, J)
                For K = 1 To N
                    a(L, K) = a(L, K) - a(J, K) * d
                    c(L, K) = c(L, K) - c(J, K) * d
                Next K
            End If
        Next L
    Next J
'
'
Inversa_2_ERR:
    Inversa_2 = (Err = 0)
    If Err <> 0 Then
        MsgBox Err.Description, vbCritical, " Inversa_2"
    End If
'
'
'
End Function
Public Function Inversa(ByRef a#(), ByRef c#()) As Boolean
'
'   Ritorna in c() l' inversa della matrice quadrata a().
'   La matrice originale a() viene sostituita dalla sua
'   scomposizione in matrici triangolari inferiore e superiore:
'
    Dim I&, J&, N&, Indx&(), d#
'
    N = UBound(a, 1)
    ReDim c(1 To N, 1 To N), b#(1 To N)
'
    For I = 1 To N              ' Set up identity matrix.
        For J = 1 To N
            c(I, J) = 0#
        Next J
        c(I, I) = 1#
    Next I
'
    ' Decompose the matrix just once:
    If ludcmp(a(), Indx(), d) Then
        ' Find inverse by columns:
        For J = 1 To N
            For I = 1 To N
                b(I) = c(I, J)
            Next I
            Call lubksb(a(), Indx(), b())
            For I = 1 To N
                c(I, J) = b(I)
            Next I
        Next J
    '
        Inversa = True
    Else
        Inversa = False
    End If
'
'
'
End Function

Public Function Determinante(ByRef a#(), ByRef det#) As Boolean
'
'   Ritorna, in det, il determinante della matrice quadrata a().
'   La matrice originale a() viene sostituita dalla sua
'   scomposizione in matrici triangolari inferiore e superiore:
'
    Dim Indx&(), J&
'
    ' This returns det as  1:
    If ludcmp(a(), Indx(), det) Then
        For J = 1 To UBound(a, 1)
            det = det * a(J, J)
        Next J
    '
        Determinante = True
    Else
        Determinante = False
    End If
'
'
'
End Function

Public Function LinEqu(ByRef a#(), ByRef b#()) As Boolean
'
'   Risolve un sistema di N equazioni lineari in N incognite.
'    a(1 To N, 1 To N): matrice dei coefficienti delle incognite.
'    b(1 To N):         vettore dei termini noti.
'   La matrice originale a() viene sostituita dalla sua
'   scomposizione in matrici triangolari inferiore e superiore.
'   Il vettore b() ritorna con le soluzioni:
'
    Dim Indx&(), d#
'
    If ludcmp(a(), Indx(), d) Then
        Call lubksb(a(), Indx(), b())
'
        LinEqu = True
    Else
        LinEqu = False
    End If
'
'
'
End Function

Private Sub elmhes(ByRef a#())
'
'   Reduction to Hessenberg form by the elimination method.
'   The real, nonsymmetric, N by N matrix a(), is replaced by an upper Hessenberg
'   matrix with identical eigenvalues.
'   Recommended, but not required, is that this routine be preceded by balanc.
'   On output, the Hessenberg matrix is in elements a(i,j) with i <= j + 1.
'   Elements with i > j + 1 are to be thought of as zero, but are returned with
'   random values.
'
    Dim I&, J&, M&, N&
    Dim x#, y#
'
    N = UBound(a, 1)
'
    For M = 2 To N - 1          ' m is called r + 1 in the text.
        x = 0#
        I = M
        For J = M To N          ' Find the pivot.
            If (Abs(a(J, M - 1)) > Abs(x)) Then
                x = a(J, M - 1)
                I = J
            End If
        Next J
'
        If (I <> M) Then        ' Interchange rows and columns.
            For J = M - 1 To N
                y = a(I, J)
                a(I, J) = a(M, J)
                a(M, J) = y
            Next J
            For J = 1 To N
                y = a(J, I)
                a(J, I) = a(J, M)
                a(J, M) = y
            Next J
        End If
'
        If (x <> 0#) Then       ' Carry out the elimination.
            For I = M + 1 To N
                y = a(I, M - 1)
                If (y <> 0#) Then
                    y = y / x
                    a(I, M - 1) = y
                    For J = M To N
                        a(I, J) = a(I, J) - y * a(M, J)
                    Next J
                    For J = 1 To N
                        a(J, M) = a(J, M) + y * a(J, I)
                    Next J
                End If
            Next I
        End If
    Next M
'
'
'
End Sub

Private Function hqr(ByRef a#(), ByRef wr#(), ByRef wi#()) As Boolean
'
'   Finds all eigenvalues of an N by N upper Hessenberg matrix a().
'   On input a() can be exactly as output from elmhes (11.5);
'   on output it is destroyed.
'   The real and imaginary parts of the eigenvalues are returned in
'   wr() and wi(), respectively.
'
    Dim I&, ITS&, J&, K&, L&, M&, N&, Nn&
    Dim anorm#, P#, Q#, r#, s#, t#, u#, v#, w#, x#, y#, z#
'
    On Error GoTo hqr_ERR
'
    N = UBound(a, 1)
    ReDim wr(1 To N), wi(1 To N)
'
    anorm = 0#      ' Compute matrix norm for possible use in locating
    For I = 1 To N  ' single small subdiagonal element.
        For J = MAX(I - 1, 1) To N
            anorm = anorm + Abs(a(I, J))
        Next J
    Next I
'
    Nn = N
    t = 0#              ' Gets changed only by an exceptional shift.
'
1   If (Nn >= 0.1) Then ' Begin search for next eigenvalue.
        ITS = 0
2       For L = Nn To 2 Step -1     ' Begin iteration: look for single small sub-
                                    ' diagonal element.
            s = Abs(a(L - 1, L - 1)) + Abs(a(L, L))
            If (s = 0#) Then s = anorm
            If ((Abs(a(L, L - 1)) + s) = s) Then GoTo 3
        Next L
        L = 1
3       x = a(Nn, Nn)
        If (L = Nn) Then            ' One root found.
            wr(Nn) = x + t
            wi(Nn) = 0#
            Nn = Nn - 1
        Else
            y = a(Nn - 1, Nn - 1)
            w = a(Nn, Nn - 1) * a(Nn - 1, Nn)
            If (L = Nn - 1) Then    ' Two roots found...
                P = 0.5 * (y - x)
                Q = P ^ 2 + w
                z = Sqr(Abs(Q))
                x = x + t
                If (Q >= 0#) Then   ' ...a real pair.
                    z = P + SIGN(z, P)
                    wr(Nn) = x + z
                    wr(Nn - 1) = wr(Nn)
                    If (z <> 0#) Then wr(Nn) = x - w / z
                    wi(Nn) = 0#
                    wi(Nn - 1) = 0#
                Else                ' ...a complex pair.
                    wr(Nn) = x + P
                    wr(Nn - 1) = wr(Nn)
                    wi(Nn) = z
                    wi(Nn - 1) = -z
                End If
                Nn = Nn - 2
'
            Else                    ' No roots found. Continue iteration.
                If (ITS = 30) Then Err.Raise 5, , "Troppe iterazioni"
'
                If (ITS = 10 Or ITS = 20) Then  'Form exceptional shift.
                    t = t + x
                    For I = 1 To Nn
                        a(I, I) = a(I, I) - x
                    Next I
                    s = Abs(a(Nn, Nn - 1)) + Abs(a(Nn - 1, Nn - 2))
                    x = 0.75 * s
                    y = x
                    w = -0.4375 * s ^ 2
                End If
'
                ITS = ITS + 1
                For M = Nn - 2 To L Step -1 ' Form shift and then look for 2 consecu-
                                            ' tive small subdiagonal elements.
                    z = a(M, M)
                    r = x - z
                    s = y - z
                    P = (r * s - w) / a(M + 1, M) + a(M, M + 1) 'Equation (11.6.23).
                    Q = a(M + 1, M + 1) - z - r - s
                    r = a(M + 2, M + 1)
                    s = Abs(P) + Abs(Q) + Abs(r)    ' Scale to prevent over ow or under ow.
                    P = P / s
                    Q = Q / s
                    r = r / s
                    If (M = L) Then GoTo 4
                    u = Abs(a(M, M - 1)) * (Abs(Q) + Abs(r))
                    v = Abs(P) * (Abs(a(M - 1, M - 1)) + Abs(z) + Abs(a(M + 1, M + 1)))
                    If (u + v = v) Then GoTo 4      ' Equation (11.6.26).
                Next M
'
4               For I = M + 2 To Nn
                    a(I, I - 2) = 0#
                    If (I <> M + 2) Then a(I, I - 3) = 0#
                Next I
                For K = M To Nn - 1                 ' Double QR step on rows l to nn
                                                    ' and columns m to nn.
                    If (K <> M) Then
                        P = a(K, K - 1)             ' Begin setup of Householder vector.
                        Q = a(K + 1, K - 1)
                        r = 0#
                        If (K <> Nn - 1) Then r = a(K + 2, K - 1)
                        x = Abs(P) + Abs(Q) + Abs(r)
                        If (x <> 0#) Then
                            P = P / x               ' Scale to prevent over ow or under ow.
                            Q = Q / x
                            r = r / x
                        End If
                    End If
                    s = SIGN(Sqr(P ^ 2 + Q ^ 2 + r ^ 2), P)
                    If (s <> 0#) Then
                        If (K = M) Then
                            If (L <> M) Then a(K, K - 1) = -a(K, K - 1)
                        Else
                            a(K, K - 1) = -s * x
                        End If
                        P = P + s                   ' Equations (11.6.24).
                        x = P / s
                        y = Q / s
                        z = r / s
                        Q = Q / P
                        r = r / P
                        For J = K To Nn             ' Row modification.
                            P = a(K, J) + Q * a(K + 1, J)
                            If (K <> Nn - 1) Then
                                P = P + r * a(K + 2, J)
                                a(K + 2, J) = a(K + 2, J) - P * z
                            End If
                            a(K + 1, J) = a(K + 1, J) - P * y
                            a(K, J) = a(K, J) - P * x
                        Next J
                        For I = L To MIN(Nn, K + 3) ' Column modification.
                            P = x * a(I, K) + y * a(I, K + 1)
                            If (K <> Nn - 1) Then
                                P = P + z * a(I, K + 2)
                                a(I, K + 2) = a(I, K + 2) - P * r
                            End If
                            a(I, K + 1) = a(I, K + 1) - P * Q
                            a(I, K) = a(I, K) - P
                        Next I
                    End If
                Next K
                GoTo 2 ' ...for next iteration on current eigenvalue.
            End If
        End If
        GoTo 1 ' ...for next eigenvalue.
    End If
'
'
hqr_ERR:
    hqr = (Err = 0)
    If Err <> 0 Then
        MsgBox Err.Description, vbCritical, " hqr"
    End If
'
'
'
End Function
Private Function MAX(ParamArray vD() As Variant) As Variant
'
'   Implementa la funzione MAX(D1, D2, ...DN) del FORTRAN:
'
    Dim J&, J1&, J2&, vDMax As Variant
'
    J1 = LBound(vD)
    J2 = UBound(vD)
    vDMax = vD(J1)
    For J = J1 + 1 To J2
        If vDMax < vD(J) Then vDMax = vD(J)
    Next J
'
    MAX = vDMax
'
'
'
End Function
Private Function MIN(ParamArray vD() As Variant) As Variant
'
'   Implementa la funzione MIN(D1, D2, ...DN) del FORTRAN:
'
    Dim J&, J1&, J2&, vDMin As Variant
'
    J1 = LBound(vD)
    J2 = UBound(vD)
    vDMin = vD(J1)
    For J = J1 + 1 To J2
        If vDMin > vD(J) Then vDMin = vD(J)
    Next J
'
    MIN = vDMin
'
'
'
End Function
Private Function SIGN(ByVal dV#, ByVal dS#) As Double
'
'   Ritorna il valore assoluto di v con il segno di s.
'   Implementa la funzione SIGN del FORTRAN:
'
    If dS < 0 Then
        SIGN = -Abs(dV)
    Else
        SIGN = Abs(dV)
    End If
'
'
'
End Function


Public Function AutoValori(a#(), ByRef wr#(), ByRef wi#()) As Boolean
'
'   Calcola tutti gli autovalori, reali e complessi della
'   matrice a(), quadrata ed a valori reali.
'   La parte reale degli autovalori viene ritornata in wr()
'   ed in wi() quella complessa.
'   La funzione ritorna False se, durante i calcoli, vengono
'   trovate condizioni anomale.
'
    balanc a()
    elmhes a()
'
    AutoValori = hqr(a(), wr(), wi())
'
'
'
End Function

Public Function Trasposta(a#()) As Double()
'
'   Ritorna la trasposta della matrice a():
'
    Dim I&, J&
    ReDim c#(1 To UBound(a, 2), 1 To UBound(a, 1))
'
    For I = 1 To UBound(a, 1)
        For J = 1 To UBound(a, 2)
            c(J, I) = a(I, J)
        Next J
    Next I
'
    Trasposta = c()
'
'
'
End Function
Public Function MDif(a#(), b#()) As Double()
'
'   Ritorna la differenza delle due matrici a() + (-b()).
'   Le due matrici devono essere dello stesso ordine:
'
    Dim I&, J&
    ReDim c#(1 To UBound(a, 1), 1 To UBound(a, 2))
'
    On Error GoTo MDif_ERR
    If (UBound(a, 1) <> UBound(b, 1)) Or (UBound(a, 2) <> UBound(b, 2)) Then Err.Raise 5
'
    For I = 1 To UBound(a, 1)
        For J = 1 To UBound(a, 2)
            c(I, J) = a(I, J) - b(I, J)
        Next J
    Next I
'
    MDif = c()
'
'
MDif_ERR:
    If Err <> 0 Then
        MsgBox "Le matrici non sono dello stesso ordine.", vbCritical, " MDif"
    End If
'
'
'
End Function
Public Function MMSc(a#(), ByVal s#) As Double()
'
'   Ritorna il prodotto della matrice a() per
'   lo scalare s:
'
    Dim I&, J&
    ReDim c#(1 To UBound(a, 1), 1 To UBound(a, 2))
'
    For I = 1 To UBound(a, 1)
        For J = 1 To UBound(a, 2)
            c(I, J) = s * a(I, J)
        Next J
    Next I
'
    MMSc = c()
'
'
'
End Function
Public Function MEgu(a#(), b#()) As Boolean
'
'   Verifica che le due matrici a() e b() siano uguali.
'   Le due matrici devono essere dello stesso ordine:
'
    Dim I&, J&
'
    On Error GoTo MEgu_ERR
    If (UBound(a, 1) <> UBound(b, 1)) Or (UBound(a, 2) <> UBound(b, 2)) Then Err.Raise 5
'
    For I = 1 To UBound(a, 1)
        For J = 1 To UBound(a, 2)
            If a(I, J) <> b(I, J) Then
                MEgu = False
                Exit Function
            End If
        Next J
    Next I
'
'
MEgu_ERR:
    MEgu = (Err = 0)
    If Err <> 0 Then
        MsgBox "Le matrici non sono dello stesso ordine.", vbCritical, " MEgu"
    End If
'
'
'
End Function
Public Function MSom(a#(), b#()) As Double()
'
'   Ritorna la somma delle due matrici a() + b().
'   Le due matrici devono essere dello stesso ordine:
'
    Dim I&, J&
    ReDim c#(1 To UBound(a, 1), 1 To UBound(a, 2))
'
    On Error GoTo MSom_ERR
    If (UBound(a, 1) <> UBound(b, 1)) Or (UBound(a, 2) <> UBound(b, 2)) Then Err.Raise 5
'
    For I = 1 To UBound(a, 1)
        For J = 1 To UBound(a, 2)
            c(I, J) = a(I, J) + b(I, J)
        Next J
    Next I
'
    MSom = c()
'
'
MSom_ERR:
    If Err <> 0 Then
        MsgBox "Le matrici non sono dello stesso ordine.", vbCritical, " MSom"
    End If
'
'
'
End Function
Public Function MMol(a#(), b#()) As Double()
'
'   Ritorna il prodotto delle due matrici a() * b().
'   Il numero di colonne di a() deve essere uguale
'   al numero di righe di b():
'
    Dim I&, J&, K&
    ReDim c#(1 To UBound(a, 1), 1 To UBound(b, 2))
'
    On Error GoTo MMol_ERR
    If UBound(a, 2) <> UBound(b, 1) Then Err.Raise 5
'
    For I = 1 To UBound(a, 1)
        For J = 1 To UBound(b, 2)
            For K = 1 To UBound(a, 2)
                c(I, J) = c(I, J) + a(I, K) * b(K, J)
            Next K
        Next J
    Next I
'
    MMol = c()
'
'
MMol_ERR:
    If Err <> 0 Then
        MsgBox "Le matrici non hanno ordini compatibili.", vbCritical, " MMol"
    End If
'
'
'
End Function


Public Function MIde(ByVal N&) As Double()
'
'   Ritorna una matrice identica (o matrice unita')
'   di ordine N con N >= 1:
'
    Dim I&
    ReDim c#(1 To N, 1 To N)
'
    For I = 1 To N
        c(I, I) = 1#
    Next I
'
    MIde = c()
'
'
'
End Function

Public Sub POWINV(a#(), ByRef x#(), ByVal Alpha#, ByRef Lambda# _
    , Optional ByVal Tol# = 0.000001, Optional ByVal MITER& = 100, Optional ByRef ITERM&)
'
'   Subroutine POWINV uses the inverse power method with scaling to
'   estimate Eigenvalues of a matrix A.  The calling program must supply
'   the matrix a(), an initial vector x() of euclidean length 1, a scalar
'   Alpha, a tolerance Tol and an integer MITER = maximum number of
'   iterations desired.  The subroutine returns when the difference of
'   two successive estimates is less than Tol in absolute value or when
'   MITER iteration have been executed.  In the first case, a flag, ITERM,
'   is set to 1 and in the second case ITERM is set to 2.  The approximate
'   Eigenvalue is returned as Lambda and an approximate Eigenvector as x().
'
'   Note di traduzione: il vettore x() viene dimensionato in questa routine
'                       e vengono assegnati i valori iniziali x(1, .., N) = 1/Sqr(N).
'
    Dim I&, J&, N&, ITR&, IERROR&
    Dim Temp#, YScale#, Estold#, Estnew#
'
    N = UBound(a, 1)
    ReDim sa#(1 To N, 1 To N), sx#(1 To N), x#(1 To N), y#(1 To N)
    For I = 1 To N
        x(I) = 1# / Sqr(CDbl(N))
    Next I
'
    ITR = 1
    For I = 1 To N
        For J = 1 To N
            sa(I, J) = a(I, J)
        Next J
    Next I
'
    ' Calculate the initial Eigenvalue approximation:
    For I = 1 To N
        sx(I) = x(I)
        a(I, I) = a(I, I) - Alpha
    Next I
'
    Call GAUS(a(), x(), y(), IERROR)
    If IERROR = 2 Then GoTo 10
'
    Temp = 0#
    YScale = 0#
    For I = 1 To N
        Temp = Temp + y(I) * sx(I)
        YScale = YScale + y(I) * y(I)
    Next I
    YScale = Sqr(YScale)
    Estold = 1# / Temp + Alpha
'
    ' Inverse power method iteration with scaling:
4   ITR = ITR + 1
    For I = 1 To N
        For J = 1 To N
            a(I, J) = sa(I, J)
        Next J
    Next I
    For I = 1 To N
        x(I) = y(I) / YScale
        sx(I) = x(I)
        a(I, I) = a(I, I) - Alpha
    Next I
'
    Call GAUS(a(), x(), y(), IERROR)
    If IERROR = 2 Then GoTo 10
'
    Temp = 0#
    YScale = 0#
    For I = 1 To N
        Temp = Temp + y(I) * sx(I)
        YScale = YScale + y(I) * y(I)
    Next I
    YScale = Sqr(YScale)
    Estnew = 1# / Temp + Alpha
'
    ' Test for termination of the inverse power method iteration:
    If (Abs(Estnew - Estold) <= Tol) Then GoTo 8
    If (ITR >= MITER) Then GoTo 9
    Estold = Estnew
    GoTo 4
'
8   ITERM = 1
    Lambda = Estnew
    For I = 1 To N              ' Aggiunta al programma originale
        x(I) = y(I) / YScale    ' per ottenere valori di scala
    Next I                      ' inferiore.
    Exit Sub
'
9   ITERM = 2
    Lambda = Estnew
    Exit Sub
'
    ' Set ITERM = 3 if the Gauss elimination step fails:
10  ITERM = 3
'
'
'
End Sub
Public Sub POWERM(a#(), ByRef x#(), ByRef Lambda# _
    , Optional ByVal Tol# = 0.000001, Optional ByVal MITER& = 100, Optional ByRef ITERM&)
'
'   Subroutine POWERM uses the power method with scaling to estimate the
'   dominant Eigenvalue of a matrix A.  The calling program must supply
'   the matrix a(), an initial vector x() of euclidean length 1, a tolerance Tol
'   and an integer MITER = maximum number of power iterations desired.
'   The subroutine returns when the difference of two successive estimates is
'   less than Tol in absolute value or when MITER iteration have been executed.
'   In the first case, a flag, ITERM, is set to 1 and in the second case ITERM
'   is set to 2.  The approximate Eigenvalue is returned as Lambda and an
'   approximate Eigenvector as x().
'
'   Note di traduzione: il vettore x() viene dimensionato in questa routine
'                       e vengono assegnati i valori iniziali x(1, .., N) = 1/Sqr(N).
'
    Dim I&, J&, N&, ITR&, IERROR&
    Dim Temp#, YScale#, Estold#, Estnew#
'
    N = UBound(a, 1)
    ReDim x#(1 To N), y#(1 To N)
    For I = 1 To N
        x(I) = 1# / Sqr(CDbl(N))
    Next I
'
    ITR = 1
'
    ' Calculate the initial Eigenvalue approximation:
    For I = 1 To N
        y(I) = 0#
        For J = 1 To N
            y(I) = y(I) + a(I, J) * x(J)
        Next J
    Next I
    Temp = 0#
    YScale = 0#
    For I = 1 To N
        Temp = Temp + y(I) * x(I)
        YScale = YScale + y(I) * y(I)
    Next I
    YScale = Sqr(YScale)
    Estold = Temp
'
    ' Power method iteration with scaling:
3   ITR = ITR + 1
    For I = 1 To N
        x(I) = y(I) / YScale
    Next I
    For I = 1 To N
        y(I) = 0#
        For J = 1 To N
            y(I) = y(I) + a(I, J) * x(J)
        Next J
    Next I
    Temp = 0#
    YScale = 0#
    For I = 1 To N
        Temp = Temp + y(I) * x(I)
        YScale = YScale + y(I) * y(I)
    Next I
    YScale = Sqr(YScale)
    Estnew = Temp
'
    ' Test for termination of the power method iteration:
    If (Abs(Estnew - Estold) <= Tol) Then GoTo 7
    If (ITR >= MITER) Then GoTo 8
    Estold = Estnew
    GoTo 3
'
7   ITERM = 1
    Lambda = Estnew
    Exit Sub
'
8   ITERM = 2
    Lambda = Estnew
'
'
'
End Sub

Public Sub GAUS(ByRef a#(), ByRef b#(), ByRef x#(), ByRef IERROR&)
'
'   Subroutine GAUS uses Gauss elimination (without pivoting) to solve
'   the system A * X = B.  The calling program must supply the matrix A
'   and the vector B. Arrays a() and b() are destroyed in GAUS.
'   The solution is returned in x() and a flag, IERROR, is set to 1 if a()
'   is non singular and is set to 2 it a() is singular.
'
    Dim I&, J&, K&, N&, IP1&, NM1&, NMK&, NP1&
    Dim Temp#, Q#
'
    N = UBound(a, 1)
    ReDim x#(1 To N)
'
    NM1 = N - 1
    For I = 1 To NM1
        ' Search for non-zero pivot element and interchange rows if necessary.
        ' If non-zero pivot element is found, set IERROR = 2 and return.
        For J = I To N
            If (a(J, I) = 0#) Then GoTo 3
            For K = I To N
                Temp = a(I, K)
                a(I, K) = a(J, K)
                a(J, K) = Temp
            Next K
            Temp = b(I)
            b(I) = b(J)
            b(J) = Temp
            GoTo 4
3
        Next J
        GoTo 8
'
        ' Eliminate the coefficients of X(I) in rows I + 1 ... N:
4       IP1 = I + 1
        For K = IP1 To N
            Q = -a(K, I) / a(I, I)
            a(K, I) = 0#
            b(K) = Q * b(I) + b(K)
            For J = IP1 To N
                a(K, J) = Q * a(I, J) + a(K, J)
            Next J
        Next K
    Next I
    If (a(N, N) = 0#) Then GoTo 8
'
    ' Backsolve the equivalent triangularized system, set IERROR = 1
    ' and return:
    x(N) = b(N) / a(N, N)
    NP1 = N + 1
    For K = 1 To NM1
        Q = 0#
        NMK = N - K
        For J = 1 To K
            Q = Q + a(NMK, NP1 - J) * x(NP1 - J)
        Next J
        x(NMK) = (b(NMK) - Q) / a(NMK, NMK)
    Next K
    IERROR = 1
    Exit Sub
'
8   IERROR = 2
'
'
'
End Sub
