Attribute VB_Name = "modEige"
'============================================================
' Descrizione.....: Routines per il calcolo degli autovalori
'                   ed autovettori (reali e complessi) di una
'                   matrice reale e completa.
' Nome dei Files..: modEige.bas
' Data............: 8/7/2001
' Versione........: 1.0 a 32 bits.
' Sistema.........: VB6 sotto Windows NT.
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://members.xoom.it/flanguasco/
'                   http://www.mclink.it/personal/MC7061/
'============================================================
'
'   Routine di ingresso:
'    rg a#(), wr#(), wi#(), Matz&, z#(), Ierr&
'     a():   matrice di cui si vogliono calcolare autovalori e autovettori.
'     wr():  vettore della parte reale degli autovalori.
'     wi():  vettore della parte immaginaria degli autovalori.
'     Matz:  passare 0 per calcolare solo gli autovalori;
'            passare 1 per calcolare anche gli autovettori.
'     z():   matrice in cui vengono ritornati gli autovettori:
'            vedere il commento iniziale a rg per il formato
'            di memorizzazione.
'     Ierr:  codice di errore ritornato 0 se non si sono incontrati
'            problemi: vedere hqr e hqr2 per gli altri valori.
'
'   Note di traduzione: l' ordine N della matrice a() viene calcolato nella
'                       routine rg come UBound(a, 1).
'                       I vettori wr(), wi(), z() vengono dimensionati in
'                       questa routine.
'                       Le matrici ed i vettori marcati ByRef nella dichiarazione
'                       iniziale delle routines vengono modificati dalle stesse.
'                       E' stata mantenuta, per quanto possibile, la struttura del
'                       programma originale: ne deriva l' uso intensivo dei GoTo
'                       (cosi' invisi ai puristi...).
'
'   Nota:   tutte le matrici ed i vettori usati da queste  routines
'           iniziano dall' indice 1: e.g. a(1 To N, 1 To N)
'
Option Explicit
Public Sub rg(ByRef a#(), ByRef wr#(), ByRef wi#(), ByVal Matz&, z#(), ByRef Ierr&)
'
'   This subroutine calls the recommended sequence of
'   subroutines from the eigensystem subroutine package (eispack)
'   to find the eigenvalues and eigenvectors (if desired)
'   of a real general matrix.
'
'    on input:
'
'      N    is the order of the matrix  a().
'
'      a()  contains the real general matrix.
'
'      Matz is an integer variable set equal to zero if
'           only eigenvalues are desired.  Otherwise it is set to
'           any non-zero integer for both eigenvalues and eigenvectors.
'
'    on output:
'
'      wr() and wi()  contain the real and imaginary parts,
'           respectively, of the eigenvalues.  Complex conjugate
'           pairs of eigenvalues appear consecutively with the
'           eigenvalue having the positive imaginary part first.
'
'      z()  contains the real and imaginary parts of the eigenvectors
'           if Matz is not zero.  If the J-th eigenvalue is real, the
'           J-th column of z() contains its eigenvector.  If the J-th
'           eigenvalue is complex with positive imaginary part, the
'           J-th and (J+1)-th columns of z() contain the real and
'           imaginary parts of its eigenvector.  The conjugate of this
'           vector is the eigenvector for the conjugate eigenvalue.
'
'      Ierr is an integer output variable set equal to an error
'           completion code described in the documentation for hqr
'           and hqr2.  The normal completion code is zero.
'
'      iv1() and fv1()  are temporary storage arrays.
'
'   Questions and comments should be directed to Burton S. Garbow,
'   Mathematics and Computer Science Div, Argonne National Laboratory
'
'   This version dated august 1983.
'
'   ------------------------------------------------------------------
'
    Dim Is1&, Is2&, N&
    N = UBound(a, 1)
    ReDim wr#(1 To N), wi#(1 To N), z#(1 To N, 1 To N), fv1#(1 To N), Iv1&(1 To N)
'
10  Call balanc(N, a(), Is1, Is2, fv1())
    Call elmhes(N, Is1, Is2, a(), Iv1())
    If (Matz <> 0) Then GoTo 20
'
'   .......... find eigenvalues only ..........
    Call hqr(N, Is1, Is2, a(), wr(), wi(), Ierr)
    GoTo 50
'
'   .......... find both eigenvalues and eigenvectors ..........
20  Call eltran(N, Is1, Is2, a(), Iv1(), z())
    Call hqr2(N, Is1, Is2, a(), wr(), wi(), z(), Ierr)
    If (Ierr <> 0) Then GoTo 50
    Call balbak(N, Is1, Is2, fv1(), N, z())
'
50
'
'
'
End Sub
Private Function MIN0(ParamArray vD() As Variant) As Long
'
'   Implementa la funzione MIN0(K1, K2, ...KN) del FORTRAN:
'
    Dim J&, J1&, J2&, vDMin As Variant
'
    J1 = LBound(vD)
    J2 = UBound(vD)
    vDMin = vD(J1)
    For J = J1 + 1 To J2
        If vDMin > vD(J) Then vDMin = vD(J)
    Next J
'
    MIN0 = CLng(vDMin)
'
'
'
End Function
Private Function SIGN(ByVal dV#, ByVal dS#) As Double
'
'   Ritorna il valore assoluto di dV con il segno di dS.
'   Implementa la funzione SIGN del FORTRAN:
'
    If dS < 0 Then
        SIGN = -Abs(dV)
    Else
        SIGN = Abs(dV)
    End If
'
'
'
End Function

Private Sub hqr(ByVal N&, ByVal Low&, ByVal Igh&, ByRef h#() _
    , ByRef wr#(), ByRef wi#(), ByRef Ierr&)
'
'   This subroutine is a translation of the algol procedure hqr,
'   num. math. 14, 219-231(1970) by Martin, Peters, and Wilkinson.
'   Handbook for Auto. Comp., vol.ii-Linear Algebra, 359-371(1971).
'
'   This subroutine finds the eigenvalues of a real
'   upper Hessenberg matrix by the qr method.
'
'    on input:
'
'      N    is the order of the matrix.
'
'      Low and Igh  are integers determined by the balancing
'           subroutine  balanc.  If balanc has not been used,
'           set Low=1, Igh=N.
'
'      h()  contains the upper Hessenberg matrix.  Information about
'           the transformations used in the reduction to Hessenberg
'           form by elmhes or orthes, if performed, is stored
'           in the remaining triangle under the Hessenberg matrix.
'
'    on output:
'
'      h()  has been destroyed.  Therefore, it must be saved
'           before calling hqr if subsequent calculation and
'           back transformation of eigenvectors is to be performed.
'
'      wr() and wi()  contain the real and imaginary parts,
'           respectively, of the eigenvalues.  The eigenvalues
'           are unordered except that complex conjugate pairs
'           of values appear consecutively with the eigenvalue
'           having the positive imaginary part first.  If an
'           error exit is made, the eigenvalues should be correct
'           for indices Ierr+1,...,N.
'
'      Ierr is set to
'        zero   for normal return,
'        J      if the limit of 30*N iterations is exhausted
'               while the J-th eigenvalue is being sought.
'
'   Questions and comments should be directed to Alan K. Cline,
'   Pleasant Valley Software, 8603 Altus Cove, Austin, TX 78759.
'   Electronic mail to cline@cs.utexas.edu.
'
'   This version dated january 1989. (for the IBM 3090vf)
'
'   ------------------------------------------------------------------
'
    Dim I&, J&, K&, L&, M&, En&, Na&, Itn&, ITS&, Enm2&
    Dim P#, Q#, r#, s#, t#, w#, x#, y#, zz#, norm#, tst1#, tst2#, foo#
    Dim notlas As Boolean
'
    Ierr = 0
    norm = 0#
'   .......... store roots isolated by balanc
'              and compute matrix norm ..........
    For J = 1 To N
        For I = 1 To MIN0(N, J + 1)
            norm = norm + Abs(h(I, J))
        Next I
    Next J
'
' "    ( prefer vector"
    For I = 1 To Low - 1
        wr(I) = h(I, I)
        wi(I) = 0#
    Next I
'
' "    ( prefer vector"
    For I = Igh + 1 To N
        wr(I) = h(I, I)
        wi(I) = 0#
    Next I
'
    En = Igh
    t = 0#
    Itn = 30 * N
'   .......... search for next eigenvalues ..........
60  If (En < Low) Then GoTo 1001
    ITS = 0
    Na = En - 1
    Enm2 = Na - 1
'   .......... look for single small sub-diagonal element
'              for L=En step -1 until Low do -- ..........
'   .......... changed lower bound to "Low+1" and removed
'              test "if (L.eq.Low) goto 100" [JMM] -- ..........
70  For L = En To Low + 1 Step -1
            s = Abs(h(L - 1, L - 1)) + Abs(h(L, L))
            If (s = 0#) Then s = norm
            tst1 = s
            tst2 = tst1 + Abs(h(L, L - 1))
            If (tst2 = tst1) Then GoTo 100
    Next L
'   .......... form shift ..........
100 x = h(En, En)
    If (L = En) Then GoTo 270
    y = h(Na, Na)
    w = h(En, Na) * h(Na, En)
    If (L = Na) Then GoTo 280
    If (Itn = 0) Then GoTo 1000
    If (ITS <> 10 And ITS <> 20) Then GoTo 130
'   .......... form exceptional shift ..........
    t = t + x
'
' "    ( prefer vector"
    For I = Low To En
        h(I, I) = h(I, I) - x
    Next I
'
    s = Abs(h(En, Na)) + Abs(h(Na, Enm2))
    x = 0.75 * s
    y = x
    w = -0.4375 * s * s
130 ITS = ITS + 1
    Itn = Itn - 1
'   .......... look for two consecutive small
'              sub-diagonal elements.
'              for M=En-2 step -1 until l do -- ..........
    For M = En - 2 To L Step -1
        zz = h(M, M)
        r = x - zz
        s = y - zz
        P = (r * s - w) / h(M + 1, M) + h(M, M + 1)
        Q = h(M + 1, M + 1) - zz - r - s
        r = h(M + 2, M + 1)
        s = Abs(P) + Abs(Q) + Abs(r)
        P = P / s
        Q = Q / s
        r = r / s
        If (M = L) Then GoTo 150
        tst1 = Abs(P) * (Abs(h(M - 1, M - 1)) + Abs(zz) + Abs(h(M + 1, M + 1)))
        tst2 = tst1 + Abs(h(M, M - 1)) * (Abs(Q) + Abs(r))
        If (tst2 = tst1) Then GoTo 150
    Next M
'
150 h(M + 2, M) = 0#
' "    ( prefer vector"
    For I = M + 3 To En
        h(I, I - 2) = 0#
        h(I, I - 3) = 0#
    Next I
'   .......... double qr step involving rows L to En and
'              columns M to En ..........
    For K = M To Na
        notlas = K <> Na
        If (K = M) Then GoTo 170
        P = h(K, K - 1)
        Q = h(K + 1, K - 1)
        r = 0#
        If (notlas) Then r = h(K + 2, K - 1)
        x = Abs(P) + Abs(Q) + Abs(r)
        If (x = 0#) Then GoTo 260
        P = P / x
        Q = Q / x
        r = r / x
170     s = SIGN(Sqr(P * P + Q * Q + r * r), P)
        If (K = M) Then GoTo 180
        h(K, K - 1) = -s * x
        GoTo 190
180     If (L <> M) Then h(K, K - 1) = -h(K, K - 1)
190      P = P + s
        x = P / s
        y = Q / s
        zz = r / s
        Q = Q / P
        r = r / P
        If (notlas) Then GoTo 225
'   .......... row modification ..........
' "    ( prefer vector"
        For J = K To En
            foo = h(K, J) + Q * h(K + 1, J)
            h(K, J) = h(K, J) - foo * x
            h(K + 1, J) = h(K + 1, J) - foo * y
        Next J
'
        J = MIN0(En, K + 3)
'   .......... column modification ..........
        For I = L To J
            foo = x * h(I, K) + y * h(I, K + 1)
            h(I, K) = h(I, K) - foo
            h(I, K + 1) = h(I, K + 1) - foo * Q
        Next I
        GoTo 255
225
'   .......... row modification ..........
' "    ( prefer vector"
        For J = K To En
            foo = h(K, J) + Q * h(K + 1, J) + r * h(K + 2, J)
            h(K, J) = h(K, J) - foo * x
            h(K + 1, J) = h(K + 1, J) - foo * y
            h(K + 2, J) = h(K + 2, J) - foo * zz
        Next J
'
        J = MIN0(En, K + 3)
'   .......... column modification ..........
        For I = L To J
            foo = x * h(I, K) + y * h(I, K + 1) + zz * h(I, K + 2)
            h(I, K) = h(I, K) - foo
            h(I, K + 1) = h(I, K + 1) - foo * Q
            h(I, K + 2) = h(I, K + 2) - foo * r
        Next I
255
'
260
    Next K
'
    GoTo 70
'   .......... one root found ..........
270 wr(En) = x + t
    wi(En) = 0#
    En = Na
    GoTo 60
'   .......... two roots found ..........
280 P = (y - x) / 2#
    Q = P * P + w
    zz = Sqr(Abs(Q))
    x = x + t
    If (Q < 0#) Then GoTo 320
'   .......... real pair ..........
    zz = P + SIGN(zz, P)
    wr(Na) = x + zz
    wr(En) = wr(Na)
    If (zz <> 0#) Then wr(En) = x - w / zz
    wi(Na) = 0#
    wi(En) = 0#
    GoTo 330
'   .......... complex pair ..........
320 wr(Na) = x + P
    wr(En) = x + P
    wi(Na) = zz
    wi(En) = -zz
330 En = Enm2
    GoTo 60
'   .......... set error -- all eigenvalues have not
'              converged after 30*N iterations ..........
1000
    Ierr = En
1001
'
'
'
End Sub

Private Sub hqr2(ByVal N&, ByVal Low&, ByVal Igh&, ByRef h#() _
    , ByRef wr#(), ByRef wi#(), ByRef z#(), ByRef Ierr&)
'
'   This subroutine is a translation of the algol procedure hqr2,
'   num. math. 16, 181-204(1970) by Peters and Wilkinson.
'   Handbook for Auto. Comp., vol.ii-Linear Algebra, 372-395(1971).
'
'   This subroutine finds the eigenvalues and eigenvectors
'   of a real upper Hessenberg matrix by the qr method.  The
'   eigenvectors of a real general matrix can also be found
'   if elmhes and eltran or orthes and ortran have
'   been used to reduce this general matrix to Hessenberg form
'   and to accumulate the similarity transformations.
'
'    on input:
'
'      N    is the order of the matrix.
'
'      Low and Igh  are integers determined by the balancing
'           subroutine balanc.  If balanc has not been used,
'           set Low=1, Igh=N.
'
'      h()  contains the upper Hessenberg matrix.
'
'      z()  contains the transformation matrix produced by eltran
'           after the reduction by elmhes, or by ortran after the
'           reduction by orthes, if performed.  If the eigenvectors
'           of the Hessenberg matrix are desired, z() must contain
'           the identity matrix.
'
'    on output:
'
'      h()  has been destroyed.
'
'      wr() and wi()  contain the real and imaginary parts,
'           respectively, of the eigenvalues.  The eigenvalues
'           are unordered except that complex conjugate pairs
'           of values appear consecutively with the eigenvalue
'           having the positive imaginary part first.  If an
'           error exit is made, the eigenvalues should be correct
'           for indices Ierr+1,...,N.
'
'      z()  contains the real and imaginary parts of the eigenvectors.
'           if the I-th eigenvalue is real, the I-th column of z()
'           contains its eigenvector.  If the I-th eigenvalue is complex
'           with positive imaginary part, the I-th and (I+1)-th
'           columns of z() contain the real and imaginary parts of its
'           eigenvector.  The eigenvectors are unnormalized.  If an
'           error exit is made, none of the eigenvectors has been found.
'
'      Ierr is set to
'        zero   for normal return,
'        J      if the limit of 30*N iterations is exhausted
'               while the J-th eigenvalue is being sought.
'
'   Calls cdiv_e for complex division.
'
'   Questions and comments should be directed to Alan K. Cline,
'   Pleasant Valley Software, 8603 Altus Cove, Austin, TX 78759.
'   Electronic mail to cline@cs.utexas.edu.
'
'   This version dated january 1989. (for the IBM 3090vf)
'
'   ------------------------------------------------------------------
'
    Dim I&, J&, K&, L&, M&, En&, Na&, Itn&, ITS&, Enm2&
    Dim P#, Q#, r#, s#, t#, w#, x#, y#, ra#, sa#
    Dim vi#, vr#, zz#, norm#, tst1#, tst2#, foo#
    Dim notlas As Boolean
'
    Ierr = 0
    norm = 0#
'   .......... store roots isolated by balanc
'              and compute matrix norm ..........
    For J = 1 To N
        For I = 1 To MIN0(N, J + 1)
            norm = norm + Abs(h(I, J))
        Next I
    Next J
'
' "    ( prefer vector"
    For I = 1 To Low - 1
        wr(I) = h(I, I)
        wi(I) = 0#
    Next I
'
' "    ( prefer vector"
    For I = Igh + 1 To N
        wr(I) = h(I, I)
        wi(I) = 0#
    Next I
'
    En = Igh
    t = 0#
    Itn = 30 * N
'   .......... search for next eigenvalues ..........
60  If (En < Low) Then GoTo 340
    ITS = 0
    Na = En - 1
    Enm2 = Na - 1
'   .......... look for single small sub-diagonal element
'              for L=En step -1 until Low do -- ..........
70  For L = En To Low + 1 Step -1
        s = Abs(h(L - 1, L - 1)) + Abs(h(L, L))
        If (s = 0#) Then s = norm
        tst1 = s
        tst2 = tst1 + Abs(h(L, L - 1))
        If (tst2 = tst1) Then GoTo 100
    Next L
'   .......... form shift ..........
100 x = h(En, En)
    If (L = En) Then GoTo 270
    y = h(Na, Na)
    w = h(En, Na) * h(Na, En)
    If (L = Na) Then GoTo 280
    If (Itn = 0) Then GoTo 1000
    If (ITS <> 10 And ITS <> 20) Then GoTo 130
'   .......... form exceptional shift ..........
    t = t + x
'
' "    ( prefer vector"
    For I = Low To En
        h(I, I) = h(I, I) - x
    Next I
'
    s = Abs(h(En, Na)) + Abs(h(Na, Enm2))
    x = 0.75 * s
    y = x
    w = -0.4375 * s * s
130 ITS = ITS + 1
    Itn = Itn - 1
'   .......... look for two consecutive small
'              sub-diagonal elements.
'              for M=Ln-2 step -1 until L do -- ..........
    For M = En - 2 To L Step -1
        zz = h(M, M)
        r = x - zz
        s = y - zz
        P = (r * s - w) / h(M + 1, M) + h(M, M + 1)
        Q = h(M + 1, M + 1) - zz - r - s
        r = h(M + 2, M + 1)
        s = Abs(P) + Abs(Q) + Abs(r)
        P = P / s
        Q = Q / s
        r = r / s
        If (M = L) Then GoTo 150
        tst1 = Abs(P) * (Abs(h(M - 1, M - 1)) + Abs(zz) + Abs(h(M + 1, M + 1)))
        tst2 = tst1 + Abs(h(M, M - 1)) * (Abs(Q) + Abs(r))
        If (tst2 = tst1) Then GoTo 150
    Next M
'
150 h(M + 2, M) = 0#
' "    ( prefer vector"
    For I = M + 3 To En
         h(I, I - 2) = 0#
         h(I, I - 3) = 0#
    Next I
'   .......... double qr step involving rows L to En and
'              columns M to En ..........
    For K = M To Na
        notlas = K <> Na
        If (K = M) Then GoTo 170
        P = h(K, K - 1)
        Q = h(K + 1, K - 1)
        r = 0#
        If (notlas) Then r = h(K + 2, K - 1)
        x = Abs(P) + Abs(Q) + Abs(r)
        If (x = 0#) Then GoTo 260
        P = P / x
        Q = Q / x
        r = r / x
170     s = SIGN(Sqr(P * P + Q * Q + r * r), P)
        If (K = M) Then GoTo 180
        h(K, K - 1) = -s * x
        GoTo 190
180     If (L <> M) Then h(K, K - 1) = -h(K, K - 1)
190     P = P + s
        x = P / s
        y = Q / s
        zz = r / s
        Q = Q / P
        r = r / P
        If (notlas) Then GoTo 225
'   .......... row modification ..........
' "    ( prefer vector"
        For J = K To N
            foo = h(K, J) + Q * h(K + 1, J)
            h(K, J) = h(K, J) - foo * x
            h(K + 1, J) = h(K + 1, J) - foo * y
        Next J
'
        J = MIN0(En, K + 3)
'   .......... column modification ..........
        For I = 1 To J
            foo = x * h(I, K) + y * h(I, K + 1)
            h(I, K) = h(I, K) - foo
            h(I, K + 1) = h(I, K + 1) - foo * Q
        Next I
'   .......... accumulate transformations ..........
        For I = Low To Igh
            foo = x * z(I, K) + y * z(I, K + 1)
            z(I, K) = z(I, K) - foo
            z(I, K + 1) = z(I, K + 1) - foo * Q
        Next I
        GoTo 255
225
'   .......... row modification ..........
' "    ( prefer vector"
        For J = K To N
            foo = h(K, J) + Q * h(K + 1, J) + r * h(K + 2, J)
            h(K, J) = h(K, J) - foo * x
            h(K + 1, J) = h(K + 1, J) - foo * y
            h(K + 2, J) = h(K + 2, J) - foo * zz
        Next J
'
        J = MIN0(En, K + 3)
'   .......... column modification ..........
        For I = 1 To J
            foo = x * h(I, K) + y * h(I, K + 1) + zz * h(I, K + 2)
            h(I, K) = h(I, K) - foo
            h(I, K + 1) = h(I, K + 1) - foo * Q
            h(I, K + 2) = h(I, K + 2) - foo * r
        Next I
'   .......... accumulate transformations ..........
        For I = Low To Igh
            foo = x * z(I, K) + y * z(I, K + 1) + zz * z(I, K + 2)
            z(I, K) = z(I, K) - foo
            z(I, K + 1) = z(I, K + 1) - foo * Q
            z(I, K + 2) = z(I, K + 2) - foo * r
        Next I
255
'
260
    Next K
'
    GoTo 70
'   .......... one root found ..........
270 h(En, En) = x + t
    wr(En) = h(En, En)
    wi(En) = 0#
    En = Na
    GoTo 60
'   .......... two roots found ..........
280 P = (y - x) / 2#
    Q = P * P + w
    zz = Sqr(Abs(Q))
    h(En, En) = x + t
    x = h(En, En)
    h(Na, Na) = y + t
    If (Q < 0#) Then GoTo 320
'   .......... real pair ..........
    zz = P + SIGN(zz, P)
    wr(Na) = x + zz
    wr(En) = wr(Na)
    If (zz <> 0#) Then wr(En) = x - w / zz
    wi(Na) = 0#
    wi(En) = 0#
    x = h(En, Na)
    s = Abs(x) + Abs(zz)
    P = x / s
    Q = zz / s
    r = Sqr(P * P + Q * Q)
    P = P / r
    Q = Q / r
'   .......... row modification ..........
' "    ( prefer vector"
    For J = Na To N
        foo = h(Na, J)
        h(Na, J) = Q * foo + P * h(En, J)
        h(En, J) = Q * h(En, J) - P * foo
    Next J
'   .......... column modification ..........
    For I = 1 To En
        foo = h(I, Na)
        h(I, Na) = Q * foo + P * h(I, En)
        h(I, En) = Q * h(I, En) - P * foo
    Next I
'   .......... accumulate transformations ..........
    For I = Low To Igh
        foo = z(I, Na)
        z(I, Na) = Q * foo + P * z(I, En)
        z(I, En) = Q * z(I, En) - P * foo
    Next I
'
    GoTo 330
'   .......... complex pair ..........
320 wr(Na) = x + P
    wr(En) = x + P
    wi(Na) = zz
    wi(En) = -zz
330 En = Enm2
    GoTo 60
'   .......... all roots found.  backsubstitute to find
'              vectors of upper triangular form ..........
340 If (norm = 0#) Then GoTo 1001
'   .......... for En=N step -1 until 1 do -- ..........
    For En = N To 1 Step -1
        P = wr(En)
        Q = wi(En)
        Na = En - 1
        'if (q) 710, 600, 800
        If Q < 0 Then
            GoTo 710
        ElseIf Q = 0 Then
            GoTo 600
        Else
            GoTo 800
        End If
'   .......... real vector ..........
600     M = En
        h(En, En) = 1#
        If (Na = 0) Then GoTo 800
'   .......... for I=En-1 step -1 until 1 do -- ..........
        For I = En - 1 To 1 Step -1
            w = h(I, I) - P
            r = 0#
'
            For J = M To En
                r = r + h(I, J) * h(J, En)
            Next J
'
            If (wi(I) >= 0#) Then GoTo 630
            zz = w
            s = r
            GoTo 700
630         M = I
            If (wi(I) <> 0#) Then GoTo 640
            t = w
            If (t <> 0#) Then GoTo 635
            tst1 = norm
            t = tst1
632         t = 0.01 * t
            tst2 = norm + t
            If (tst2 > tst1) Then GoTo 632
635         h(I, En) = -r / t
            GoTo 680
'   .......... solve real equations ..........
640         x = h(I, I + 1)
            y = h(I + 1, I)
            Q = (wr(I) - P) * (wr(I) - P) + wi(I) * wi(I)
            t = (x * s - zz * r) / Q
            h(I, En) = t
            If (Abs(x) <= Abs(zz)) Then GoTo 650
            h(I + 1, En) = (-r - w * t) / x
            GoTo 680
650         h(I + 1, En) = (-s - y * t) / zz
'
'   .......... overflow control ..........
680         t = Abs(h(I, En))
            If (t = 0#) Then GoTo 700
            tst1 = t
            tst2 = tst1 + 1# / tst1
            If (tst2 > tst1) Then GoTo 700
            For J = I To En
                h(J, En) = h(J, En) / t
            Next J
'
700
        Next I
'   .......... end real vector ..........
        GoTo 800
'   .......... complex vector ..........
710     M = Na
'   .......... last vector component chosen imaginary so that
'              eigenvector matrix is triangular ..........
        If (Abs(h(En, Na)) <= Abs(h(Na, En))) Then GoTo 720
        h(Na, Na) = Q / h(En, Na)
        h(Na, En) = -(h(En, En) - P) / h(En, Na)
        GoTo 730
720     Call CDiv_e(0#, -h(Na, En), h(Na, Na) - P, Q, h(Na, Na), h(Na, En))
730     h(En, Na) = 0#
        h(En, En) = 1#
        Enm2 = Na - 1
        If (Enm2 = 0) Then GoTo 800
'   .......... for I=En-2 step -1 until 1 do -- ..........
        For I = En - 2 To 1 Step -1
            w = h(I, I) - P
            ra = 0#
            sa = 0#
'
            For J = M To En
                ra = ra + h(I, J) * h(J, Na)
                sa = sa + h(I, J) * h(J, En)
            Next J
'
            If (wi(I) >= 0#) Then GoTo 770
            zz = w
            r = ra
            s = sa
            GoTo 795
770         M = I
            If (wi(I) <> 0#) Then GoTo 780
            Call CDiv_e(-ra, -sa, w, Q, h(I, Na), h(I, En))
            GoTo 790
'   .......... solve complex equations ..........
780         x = h(I, I + 1)
            y = h(I + 1, I)
            vr = (wr(I) - P) * (wr(I) - P) + wi(I) * wi(I) - Q * Q
            vi = (wr(I) - P) * 2# * Q
            If (vr <> 0# Or vi <> 0#) Then GoTo 784
            tst1 = norm * (Abs(w) + Abs(Q) + Abs(x) + Abs(y) + Abs(zz))
            vr = tst1
783         vr = 0.01 * vr
            tst2 = tst1 + vr
            If (tst2 > tst1) Then GoTo 783
784         Call CDiv_e(x * r - zz * ra + Q * sa, x * s - zz * sa - Q * ra _
                      , vr, vi, h(I, Na), h(I, En))
            If (Abs(x) <= Abs(zz) + Abs(Q)) Then GoTo 785
            h(I + 1, Na) = (-ra - w * h(I, Na) + Q * h(I, En)) / x
            h(I + 1, En) = (-sa - w * h(I, En) - Q * h(I, Na)) / x
            GoTo 790
785         Call CDiv_e(-r - y * h(I, Na), -s - y * h(I, En), zz _
                      , Q, h(I + 1, Na), h(I + 1, En))
'
'   .......... overflow control ..........
790         t = DMAX1(Abs(h(I, Na)), Abs(h(I, En)))
            If (t = 0#) Then GoTo 795
            tst1 = t
            tst2 = tst1 + 1# / tst1
            If (tst2 > tst1) Then GoTo 795
            For J = I To En
                h(J, Na) = h(J, Na) / t
                h(J, En) = h(J, En) / t
            Next J
'
795
        Next I
'   .......... end complex vector ..........
800
    Next En
'   .......... end back substitution.
'              vectors of isolated roots ..........
    For I = 1 To Low - 1
' "    ( prefer vector"
        For J = I To N
            z(I, J) = h(I, J)
        Next J
    Next I
'
    For I = Igh + 1 To N
' "    ( prefer vector"
        For J = I To N
            z(I, J) = h(I, J)
        Next J
    Next I
'
'   .......... multiply by transformation matrix to give
'              vectors of original full matrix.
'              for J=N step -1 until Low do -- ..........
    For J = N To Low Step -1
        For I = Low To Igh
            zz = 0#
' "    ( prefer vector"
            For K = Low To MIN0(J, Igh)
                zz = zz + z(I, K) * h(K, J)
            Next K
                z(I, J) = zz
        Next I
    Next J
'
    GoTo 1001
'   .......... set error -- all eigenvalues have not
'              converged after 30*N iterations ..........
1000
    Ierr = En
1001
'
'
'
End Sub
Private Function DMAX1(ParamArray vD() As Variant) As Double
'
'   Implementa la funzione DMAX1(D1, D2, ...DN) del FORTRAN:
'
    Dim J&, J1&, J2&, vDMax As Variant
'
    J1 = LBound(vD)
    J2 = UBound(vD)
    vDMax = vD(J1)
    For J = J1 + 1 To J2
        If vDMax < vD(J) Then vDMax = vD(J)
    Next J
'
    DMAX1 = CDbl(vDMax)
'
'
'
End Function

Private Sub balbak(ByVal N&, ByVal Low&, ByVal Igh&, sscale#(), ByVal M&, ByRef z#())
'
'   This subroutine is a translation of the algol procedure balbak,
'   num. math. 13, 293-304(1969) by Parlett and Reinsch.
'   Handbook for Auto. Comp., vol.ii-Linear Algebra, 315-326(1971).
'
'   This subroutine forms the eigenvectors of a real general
'   matrix by back transforming those of the corresponding
'   balanced matrix determined by  balanc.
'
'    on input:
'
'      N    is the order of the matrix.
'
'      Low and Igh  are integers determined by  balanc.
'
'      sscale() contains information determining the permutations
'           and scaling factors used by balanc.
'
'      M    is the number of columns of z() to be back transformed.
'
'      z()  contains the real and imaginary parts of the eigen-
'           vectors to be back transformed in its first m columns.
'
'    on output:
'
'      z()  contains the real and imaginary parts of the
'           transformed eigenvectors in its first M columns.
'
'   Questions and comments should be directed to Burton S. Garbow,
'   Mathematics and Computer Science Div, Argonne National Laboratory.
'
'   This version dated august 1983.
'
'   ------------------------------------------------------------------
'
    Dim I&, J&, K&, II&
    Dim s#
'
    If (M = 0) Then GoTo 200
    If (Igh = Low) Then GoTo 120
'
    For I = Low To Igh
        s = sscale(I)
'   .......... left hand eigenvectors are back transformed
'              if the foregoing statement is replaced by
'              s=1.0d0/sscale(I). ..........
        For J = 1 To M
            z(I, J) = z(I, J) * s
        Next J
'
    Next I
'   ......... for I=Low-1 step -1 until 1,
'             Igh+1 step 1 until N do -- ..........
120 For II = 1 To N
        I = II
        If (I >= Low And I <= Igh) Then GoTo 140
        If (I < Low) Then I = Low - II
        K = sscale(I)
        If (K = I) Then GoTo 140
'
        For J = 1 To M
            s = z(I, J)
            z(I, J) = z(K, J)
            z(K, J) = s
        Next J
'
140
    Next II
'
200
'
'
'
End Sub

Private Sub balanc(ByVal N&, ByRef a#(), ByRef Low&, ByRef Igh&, sscale#())
'
'   This subroutine is a translation of the algol procedure balance,
'   num. math. 13, 293-304(1969) by Parlett and Reinsch.
'   Handbook for Auto. Comp., vol.ii-Linear Algebra, 315-326(1971).
'
'   This subroutine balances a real matrix and isolates
'   eigenvalues whenever possible.
'
'    on input:
'
'      N    is the order of the matrix.
'
'      a()  contains the input matrix to be balanced.
'
'    on output:
'
'      a()  contains the balanced matrix.
'
'      Low and Igh  are two integers such that a(I,J)
'           is equal to zero if
'            (1) I is greater than J and
'            (2) J=1,...,Low-1 or I=Igh+1,...,N.
'
'      sscale  contains information determining the
'           permutations and scaling factors used.
'
'   Suppose that the principal submatrix in rows Low through Igh
'   has been balanced, that p(J) denotes the index interchanged
'   with J during the permutation step, and that the elements
'   of the diagonal matrix used are denoted by d(I,J).  Then
'      sscale(J) = p(J),    for J = 1,...,Low-1
'                = d(J,J),      J = Low,...,Igh
'                = p(J)         J = Igh+1,...,N.
'   the order in which the interchanges are made is N to Igh+1,
'   then 1 to Low-1.
'
'   Note that 1 is returned for Igh if Igh is zero formally.
'
'   The algol procedure exc contained in balance appears in
'   balanc in line.  (note that the algol roles of identifiers
'   K,L have been reversed.)
'
'   Questions and comments should be directed to Alan K. Cline,
'   Pleasant Valley Software, 8603 Altus Cove, Austin, TX 78759.
'   Electronic mail to cline@cs.utexas.edu.
'
'   This version dated january 1989. (for the IBM 3090vf)
'
'   ------------------------------------------------------------------
    Dim I&, J&, K&, L&, M&, Iexc&
    Dim c#, f#, g#, r#, s#, b2#, RADIX#
    Dim noconv As Boolean
'
    RADIX = 16#
'
    b2 = RADIX * RADIX
    K = 1
    L = N
    GoTo 100
'   .......... in-line procedure for row and
'            column exchange ..........
20  sscale(M) = J
    If (J = M) Then GoTo 50
'
    For I = 1 To L
        f = a(I, J)
        a(I, J) = a(I, M)
        a(I, M) = f
    Next
'
    For I = K To N
        f = a(J, I)
        a(J, I) = a(M, I)
        a(M, I) = f
    Next I
'
50
    On Iexc GoTo 80, 130
'   .......... search for rows isolating an eigenvalue
'              and push them down ..........
80  If (L = 1) Then GoTo 280
      L = L - 1
'   .......... for J=l step -1 until 1 do -- ..........
100 For J = L To 1 Step -1
        For I = 1 To L
            If (I = J) Then GoTo 110
            If (a(J, I) <> 0#) Then GoTo 120
110
        Next I
'
         M = L
         Iexc = 1
         GoTo 20
120
    Next J
'
    GoTo 140
'   .......... search for columns isolating an eigenvalue
'              and push them left ..........
130 K = K + 1
'
140 For J = K To L
'
        For I = K To L
            If (I = J) Then GoTo 150
            If (a(I, J) <> 0#) Then GoTo 170
150
        Next I
'
         M = K
         Iexc = 2
         GoTo 20
170
    Next J
'   .......... now balance the submatrix in rows K to L ..........
    For I = K To L
        sscale(I) = 1#
    Next I
'   .......... iterative loop for norm reduction ..........
190 noconv = False
'
    For I = K To L
        c = 0#
        r = 0#
'
        For J = K To L
            If (J = I) Then GoTo 200
            c = c + Abs(a(J, I))
            r = r + Abs(a(I, J))
200
        Next J
'   .......... guard against zero c or r due to underflow ..........
        If (c = 0# Or r = 0#) Then GoTo 270
        g = r / RADIX
        f = 1#
        s = c + r
210     If (c >= g) Then GoTo 220
        f = f * RADIX
        c = c * b2
        GoTo 210
220     g = r * RADIX
230     If (c < g) Then GoTo 240
        f = f / RADIX
        c = c / b2
        GoTo 230
'   .......... now balance ..........
240     If ((c + r) / f >= 0.95 * s) Then GoTo 270
        g = 1# / f
        sscale(I) = sscale(I) * f
        noconv = True
'
        For J = K To N
250         a(I, J) = a(I, J) * g
        Next J
'
        For J = 1 To L
260         a(J, I) = a(J, I) * f
        Next J
'
270
    Next I
'
    If (noconv) Then GoTo 190
'
280 Low = K
    Igh = L
'
'
'
End Sub

Private Sub eltran(ByVal N&, ByVal Low&, ByVal Igh&, a#(), Iint&(), ByRef z#())
'
'   This subroutine is a translation of the algol procedure elmtrans,
'   num. math. 16, 181-204(1970) by Peters and Wilkinson.
'   Handbook for Auto. Comp., vol.ii-Linear Algebra, 372-395(1971).
'
'   This subroutine accumulates the stabilized elementary
'   similarity transformations used in the reduction of a
'   real general matrix to upper Hessenberg form by elmhes.
'
'    on input:
'
'      N    is the order of the matrix.
'
'      Low and Igh  are integers determined by the balancing
'           subroutine balanc.  If balanc has not been used,
'           set Low=1, Igh=N.
'
'      a()  contains the multipliers which were used in the
'           reduction by elmhes in its lower triangle
'           below the subdiagonal.
'
'      Iint() contains information on the rows and columns
'           interchanged in the reduction by elmhes.
'           Only elements Low through Igh are used.
'
'    on output:
'
'      z()  contains the transformation matrix produced in the
'           reduction by elmhes.
'
'   Questions and comments should be directed to Alan K. Cline,
'   Pleasant Valley Software, 8603 Altus Cove, Austin, TX 78759.
'   Electronic mail to cline@cs.utexas.edu.
'
'   This version dated january 1989. (for the IBM 3090vf)
'
'   ------------------------------------------------------------------
'
    Dim I&, J&, MP&
'
'   .......... initialize z to identity matrix ..........
    For J = 1 To N
'
        For I = 1 To N
            z(I, J) = 0#
        Next I
'
        z(J, J) = 1#
    Next J
'
    For MP = Igh - 1 To Low + 1 Step -1
        For I = MP + 1 To Igh
            z(I, MP) = a(I, MP - 1)
        Next I
'
        I = Iint(MP)
'
        For J = MP To Igh
            z(MP, J) = z(I, J)
            z(I, J) = 0#
        Next J
'
        z(I, MP) = 1#
    Next MP
'
'
'
End Sub

Private Sub CDiv_e(ByVal ar#, ByVal ai#, ByVal br#, ByVal bi#, ByRef cr#, ByRef ci#)
'
'   Complex division, (cr,ci) = (ar,ai)/(br,bi)
'
    Dim s#, ars#, ais#, brs#, bis#
'
    s = Abs(br) + Abs(bi)
    ars = ar / s
    ais = ai / s
    brs = br / s
    bis = bi / s
    s = brs * brs + bis * bis
    cr = (ars * brs + ais * bis) / s
    ci = (ais * brs - ars * bis) / s
'
'
'
End Sub

Private Sub elmhes(ByVal N&, ByVal Low&, ByVal Igh&, ByRef a#(), ByRef Iint&())
'
'   This subroutine is a translation of the algol procedure elmhes,
'   num. math. 12, 349-368(1968) by Martin and Wilkinson.
'   Handbook for Auto. Comp., vol.ii-Linear Algebra, 339-358(1971).
'
'   Given a real general matrix, this subroutine
'   reduces a submatrix situated in rows and columns
'   Low through Igh to upper Hessenberg form by
'   stabilized elementary similarity transformations.
'
'    on input:
'
'      N    is the order of the matrix.
'
'      Low and Igh  are integers determined by the balancing
'           subroutine  balanc.  If  balanc  has not been used,
'           set Low=1, Igh=N.
'
'      a()  contains the input matrix.
'
'    on output:
'
'      a()  contains the Hessenberg matrix.  The multipliers
'           which were used in the reduction are stored in the
'           remaining triangle under the Hessenberg matrix.
'
'      Iint() contains information on the rows and columns
'           interchanged in the reduction.
'           Only elements Low through Igh are used.
'
'   Questions and comments should be directed to Alan K. Cline,
'   Pleasant Valley Software, 8603 Altus Cove, Austin, TX 78759.
'   Electronic mail to cline@cs.utexas.edu.
'
'   This version dated january 1989. (for the IBM 3090vf)
'
'   ------------------------------------------------------------------
'
    Dim I&, J&, M&, MM1&, MP1&, IP&
    Dim x#, y#
    ReDim Iint&(1 To Igh)
'
    For M = Low + 1 To Igh - 1
        MM1 = M - 1
        x = 0#
        IP = M
'
        For J = M To Igh
            If (Abs(a(J, MM1)) <= Abs(x)) Then GoTo 100
            x = a(J, MM1)
            IP = J
100
        Next J
'
        Iint(M) = IP
        If (x = 0#) Then GoTo 180
'
'   .......... interchange columns of a ..........
        For J = 1 To Igh
            y = a(J, IP)
            a(J, IP) = a(J, M)
            a(J, M) = y
        Next J
'
        y = a(IP, MM1)
        a(IP, MM1) = a(M, MM1)
        a(M, MM1) = y
'
        MP1 = M + 1
'
        For I = MP1 To Igh
            a(I, MM1) = a(I, MM1) / x
        Next I
'
'   To avoid passing through the address space of a two times
'   (as would be done if the LHS and RHS updates were performed
'   one after the other) these updates are interleaved.
'   Note that the inner loop, with label 138, must be executed
'   for J=M,N, while the other iner loop, 139, is executed only
'   for J=M+1,Igh.)  To accomplish this, the outer loop (on J=M,N below)
'   must be broken into three separate cases: J=M | J=M+1,Igh | J=Igh+1,N.
'
        y = a(IP, M)
        a(IP, M) = a(M, M)
        a(M, M) = y
        y = -y
        For I = MP1 To Igh
            a(I, M) = a(I, M) + y * a(I, MM1)
        Next I
'
        For J = MP1 To Igh
            y = a(IP, J)
            a(IP, J) = a(M, J)
            a(M, J) = y
            y = -y
            For I = MP1 To Igh
138             a(I, J) = a(I, J) + y * a(I, MM1)
            Next I
' "    ( ignore recrdeps"
            For I = 1 To Igh
139             a(I, M) = a(I, M) + a(J, MM1) * a(I, J)
            Next I
        Next J
'
        For J = Igh + 1 To N
            y = a(IP, J)
            a(IP, J) = a(M, J)
            a(M, J) = y
            y = -y
            For I = MP1 To Igh
                a(I, J) = a(I, J) + y * a(I, MM1)
            Next I
        Next J
'
180
    Next M
'
'
'
End Sub
